import sys

import collections

from ..._compat import iteritems, itervalues, OrderedDict, reraise
from ...ctx import context as slash_context
from ...exception_handling import handling_exceptions
from ...exceptions import CyclicFixtureDependency, UnresolvedFixtureStore
from ...utils.python import getargspec
from .fixture import Fixture
from .namespace import Namespace
from .parameters import Parametrization, get_parametrizations
from .utils import get_scope_by_name, nofixtures
from .variation import VariationFactory
from .active_fixture import ActiveFixture

class FixtureStore(object):

    def __init__(self):
        super(FixtureStore, self).__init__()
        self._namespaces = [Namespace(self)]
        self._unresolved_fixture_ids = set()
        self._fixtures_by_id = {}
        self._fixtures_by_fixture_info = {}
        self._active_fixtures_by_scope = collections.defaultdict(OrderedDict)
        self._computing = set()
        self._all_needed_parametrization_ids_by_fixture_id = {}

    def get_active_fixture(self, fixture):
        return self._active_fixtures_by_scope[fixture.info.scope].get(fixture.info.id)

    def call_with_fixtures(self, test_func, namespace, is_method):

        if not nofixtures.is_marked(test_func):
            arg_names = self.get_required_fixture_names(test_func, is_method=is_method)
            kwargs = self.get_fixture_dict(arg_names, namespace)
        else:
            kwargs = {}

        return test_func(**kwargs)

    def get_required_fixture_names(self, test_func, is_method):
        skip_names = set(name for p in get_parametrizations(test_func) for name in p.names)
        arg_names = [name for name in getargspec(test_func).args if name not in skip_names]
        if is_method:
            arg_names = arg_names[1:]
        return arg_names

    def get_required_fixture_objects(self, test_func, namespace, is_method):
        names = self.get_required_fixture_names(test_func, is_method=is_method)
        return set(itervalues(self.get_fixture_dict(names, namespace=namespace, get_values=False)))

    def __iter__(self):
        return itervalues(self._fixtures_by_id)

    def push_namespace(self):
        self._namespaces.append(Namespace(self, parent=self._namespaces[-1]))

    def pop_namespace(self):
        return self._namespaces.pop(-1)

    def get_current_namespace(self):
        return self._namespaces[-1]

    def get_all_needed_parametrization_ids(self, fixtureobj):
        if self._unresolved_fixture_ids:
            raise UnresolvedFixtureStore()
        if isinstance(fixtureobj, Parametrization):
            return frozenset([fixtureobj.info.id])
        returned = self._all_needed_parametrization_ids_by_fixture_id.get(fixtureobj.info.id)
        if returned is None:
            returned = self._compute_all_needed_parametrization_ids(fixtureobj)
            self._all_needed_parametrization_ids_by_fixture_id[fixtureobj.info.id] = returned
        return returned

    def iter_autouse_fixtures_in_namespace(self, namespace=None):
        if namespace is None:
            namespace = self.get_current_namespace()
        for fixture in namespace.iter_fixtures():
            if fixture.info.autouse:
                yield fixture

    def activate_autouse_fixtures_in_namespace(self, namespace):
        for fixture in self.iter_autouse_fixtures_in_namespace(namespace):
            _ = self.get_fixture_value(fixture)

    def _compute_all_needed_parametrization_ids(self, fixtureobj):
        stack = [(fixtureobj.info.id, [fixtureobj.info.id], set([fixtureobj.info.id]))]
        returned = set()
        while stack:
            fixture_id, path, visited = stack.pop()
            if fixture_id in self._all_needed_parametrization_ids_by_fixture_id:
                returned.update(self._all_needed_parametrization_ids_by_fixture_id[fixture_id])
                continue
            fixture = self._fixtures_by_id[fixture_id]
            if fixture.parametrization_ids:
                returned.update(fixture.parametrization_ids)
            if fixture.fixture_kwargs:
                for needed_id in itervalues(fixture.fixture_kwargs):
                    if needed_id in visited:
                        self._raise_cyclic_dependency_error(fixtureobj, path, needed_id)
                    stack.append((needed_id, path + [needed_id], visited | set([needed_id])))
        return frozenset(returned)

    def _raise_cyclic_dependency_error(self, fixtureobj, path, new_id):
        raise CyclicFixtureDependency(
            'Cyclic fixture dependency detected in {0}: {1}'.format(
                fixtureobj.info.func.__code__.co_filename,
                ' -> '.join(self._fixtures_by_id[f_id].info.name
                            for f_id in path + [new_id])))

    def push_scope(self, scope):
        scope = get_scope_by_name(scope)

    def pop_scope(self, scope, in_failure, in_interruption): # pylint: disable=unused-argument
        if in_interruption:
            return
        scope = get_scope_by_name(scope)
        for s, active_fixtures in iteritems(self._active_fixtures_by_scope):
            if s <= scope:
                for active_fixture in list(active_fixtures.values())[::-1]:
                    with handling_exceptions(swallow=True):
                        self._deactivate_fixture(active_fixture.fixture)
                assert not active_fixtures

    def ensure_known_parametrization(self, parametrization):
        if parametrization.info.id not in self._fixtures_by_id:
            self._fixtures_by_id[parametrization.info.id] = parametrization

    def add_fixtures_from_dict(self, d):
        for name, thing in iteritems(d):
            fixture_info = getattr(thing, '__slash_fixture__', None)
            if fixture_info is None:
                continue
            self.get_current_namespace().add_name(
                name, self.add_fixture(thing).__slash_fixture__.id)

    def add_fixture(self, fixture_func):
        fixture_info = fixture_func.__slash_fixture__
        existing_fixture = self._fixtures_by_id.get(fixture_info.id)
        if existing_fixture is not None:
            return existing_fixture.fixture_func
        fixture_object = Fixture(self, fixture_func)
        current_namespace = self._namespaces[-1]
        current_namespace.add_name(fixture_info.name, fixture_info.id)
        self.register_fixture_id(fixture_object)
        return fixture_func

    def register_fixture_id(self, f):
        assert f.info.id not in self._fixtures_by_id
        self._fixtures_by_id[f.info.id] = f
        self._unresolved_fixture_ids.add(f.info.id)

    def get_fixture_by_name(self, name):
        return self._namespaces[-1].get_fixture_by_name(name)

    def get_fixture_by_id(self, fixture_id):
        return self._fixtures_by_id[fixture_id]

    def get_fixture_dict(self, required_names, namespace=None, get_values=True, skip_names=frozenset()):
        returned = {}

        if namespace is None:
            namespace = self.get_current_namespace()

        for required_name in required_names:
            if required_name in skip_names:
                continue
            fixture = namespace.get_fixture_by_name(required_name)
            if get_values:
                fixture = self.get_fixture_value(fixture, name=required_name)
            returned[required_name] = fixture
        return returned

    def get_fixture_value(self, fixture, name=None):
        if name is None:
            name = fixture.info.name

        value = self._fill_fixture_value(name, fixture)
        return value

    def iter_parametrization_variations(self, fixture_ids=(), funcs=(), methods=()):

        variation_factory = VariationFactory(self)
        for fixture_id in fixture_ids:
            variation_factory.add_needed_fixture_id(fixture_id)

        for func in funcs:
            variation_factory.add_needed_fixtures_from_function(func)

        for method in methods:
            variation_factory.add_needed_fixtures_from_method(method)

        return variation_factory.iter_variations()

    def _fill_fixture_value(self, name, fixture):
        if fixture.info.id in self._computing:
            raise CyclicFixtureDependency(
                'Fixture {0!r} is a part of a dependency cycle!'.format(name))

        active_fixture = self.get_active_fixture(fixture)

        if active_fixture is not None:
            return active_fixture.value

        self._computing.add(fixture.info.id)
        try:
            fixture_value = self._activate_fixture(fixture)
        except:
            exc_info = sys.exc_info()
            self._deactivate_fixture(fixture)
            reraise(*exc_info)
        finally:
            self._computing.discard(fixture.info.id)

        return fixture_value

    def _activate_fixture(self, fixture):
        active_fixture = ActiveFixture(fixture)

        kwargs = {}

        if fixture.fixture_kwargs is None:
            raise UnresolvedFixtureStore('Fixture {0} is unresolved!'.format(fixture.info.name))

        for required_name, fixture_id in iteritems(fixture.fixture_kwargs):
            kwargs[required_name] = self._fill_fixture_value(
                required_name, self.get_fixture_by_id(fixture_id))


        assert fixture.info.id not in self._active_fixtures_by_scope[fixture.info.scope]
        self._active_fixtures_by_scope[fixture.info.scope][fixture.info.id] = active_fixture
        prev_context_fixture = slash_context.fixture
        slash_context.fixture = active_fixture
        try:
            returned = active_fixture.value = fixture.get_value(kwargs, active_fixture)
        finally:
            slash_context.fixture = prev_context_fixture
        return returned

    def _deactivate_fixture(self, fixture):
        # in most cases it will be the last active fixture in its scope
        active = self._active_fixtures_by_scope[fixture.info.scope].pop(fixture.info.id, None)
        if active is not None:
            active.do_cleanups()

    def resolve(self):
        while self._unresolved_fixture_ids:
            fixture = self._fixtures_by_id[self._unresolved_fixture_ids.pop()]
            fixture.resolve(self)



import functools

from sentinels import NOTHING

from .._compat import iteritems

_TAGS_NAME = '__slash_tags__'


def tag_test(test, tag_name, tag_value):
    tags = get_tags(test)
    if tags is NO_TAGS:
        tags = Tags()
        setattr(test, _TAGS_NAME, tags)
    assert tag_name not in tags
    tags[tag_name] = tag_value
    return test


def tag(tag_name, tag_value=NOTHING):
    """Decorator for tagging tests
    """
    return functools.partial(tag_test, tag_name=tag_name, tag_value=tag_value)


def get_tags(test):
    return getattr(test, _TAGS_NAME, NO_TAGS)


class Tags(object):

    def __init__(self, tags=None):
        super(Tags, self).__init__()
        if tags is None:
            tags = {}
        self._tags = tags

    def __setitem__(self, tag_name, value):
        self._tags[tag_name] = value

    def __getitem__(self, tag_name):
        return self._tags[tag_name]

    def __contains__(self, tag_name):
        return tag_name in self._tags

    has_tag = __contains__

    def __add__(self, other):
        if other is NO_TAGS:
            return self
        new_tags = self._tags.copy()
        new_tags.update(other._tags) # pylint: disable=protected-access
        return Tags(new_tags)

    def matches_pattern(self, pattern):
        if '=' in pattern:
            key, predicate = pattern.split('=', 1)
            value = self._tags.get(key, NOTHING)
            return value is not NOTHING and str(value) == predicate

        for key, value in iteritems(self._tags):
            if pattern in key:
                return True
        return False


class _NoTags(object):

    def __contains__(self, tag_name):
        return False

    def __add__(self, other):
        return other

    has_tag = __contains__

    def matches_pattern(self, pattern): # pylint: disable=unused-argument
        return False

NO_TAGS = _NoTags()



import copy
import itertools

import logbook

import slash

from .validation import get_test_id_from_test_address

_timestamps = itertools.count()

_logger = logbook.Logger(__name__)


class SlashRunResult(object):

    _console_output = None

    def __init__(self, report_stream):
        super(SlashRunResult, self).__init__()
        self._report_stream = report_stream
        self.exit_code = 0
        self.error_message = None
        self.session = None
        self.tracker = Tracker()

    @property
    def events(self):
        return self.tracker.events

    def ok(self):
        return self.session.results.is_success()

    def get_console_output(self):
        if self._console_output is None:
            self._console_output = self._report_stream.getvalue()
        return self._console_output

    def get_all_results_for_test(self, test):
        returned = []
        for result in self.session.results:
            if get_test_id_from_test_address(result.test_metadata.address) == test.id:
                returned.append(result)
        return returned

    def __getitem__(self, test):
        matching = self.get_all_results_for_test(test)
        assert len(matching) == 1
        return matching[0]

    def __repr__(self):
        return '<Summary ({0})>'.format(self.session.results)


class Tracker(object):

    def __init__(self):
        super(Tracker, self).__init__()
        self.events = Events()
        self.active_fixtures = {}

    def notify_fixture_start(self, f_id, value):
        _logger.debug('started fixture {0}', f_id)
        assert f_id not in self.active_fixtures
        self.active_fixtures[f_id] = value

    def notify_fixture_end(self, f_id):
        _logger.debug('ended fixture {0}', f_id)
        self.active_fixtures.pop(f_id)

    def notify_parameter_value(self, p_id, value):
        slash.context.result.data.setdefault('param_values', {})[p_id] = value

    def get_fixture_memento(self):
        return copy.deepcopy(self.active_fixtures)


class Event(object):

    def __init__(self, args):
        super(Event, self).__init__()
        self.args = args
        self.timestamp = next(_timestamps)

    def is_before(self, other_event):
        return self.timestamp < other_event.timestamp


class Events(object):

    def __init__(self):
        super(Events, self).__init__()
        self._events = []
        self._events_by_args = {}

    def add(self, *args):
        self._events.append(Event(args))
        self._events_by_args[args] = self._events[-1]

    def assert_consecutive(self, evts):
        seq = [self[x] for x in evts]
        timestamps = [e.timestamp for e in seq]
        begin = seq[0].timestamp
        assert timestamps == [begin + i for i in range(len(seq))]

    def __getitem__(self, args):
        return self._events_by_args[self._normalize_key(args)]

    def __contains__(self, args):
        return self._normalize_key(args) in self._events_by_args
        return self.has_event(*args)

    def has_event(self, *args):
        return args in self

    def _normalize_key(self, args):
        if not isinstance(args, tuple):
            args = (args,)
        return args

    def __repr__(self):
        return '<Events>'




import gc

import slash
import pytest

from .utils import TestCase
from slash import Session
from slash.core.result import Result
from slash.core.result import SessionResults, GlobalResult

from .utils import run_tests_assert_success



@pytest.mark.parametrize('use_error', [True, False])
def test_result_add_exception_multiple_times(result, use_error):
    second_result = type(result)()
    second_result.mark_started()
    try:
        if use_error:
            1 / 0
        else:
            assert 1 + 1 == 3
    except:
        for i in range(3):
            result.add_exception()
        second_result.add_exception()

    assert result.is_error() == use_error
    assert result.is_failure() == (not use_error)
    assert len(result.get_errors() if use_error else result.get_failures()) == 1
    assert second_result.is_success()


def test_result_summary_some_not_run(suite):
    suite[2].add_decorator('slash.requires(False)')
    suite[2].expect_skip()
    results = suite.run().session.results
    assert results.is_success(allow_skips=True)


def test_result_summary(suite):

    suite[2].when_run.fail()
    suite[3].when_run.raise_exception()
    suite[4].when_run.raise_exception()
    suite[5].when_run.skip()

    results = suite.run().session.results

    assert results.get_num_errors() == 2
    assert results.get_num_failures() == 1
    assert results.get_num_skipped() == 1
    assert results.get_num_successful() == len(suite) - 4
    assert results.get_num_not_run() == 0


def test_result_not_run(suite, suite_test, is_last_test):
    suite_test.when_run.fail()

    for test in suite.iter_all_after(suite_test, assert_has_more=not is_last_test):
        test.expect_not_run()

    summary = suite.run(additional_args=['-x'])

    num_not_run = summary.session.results.get_num_not_run()
    if is_last_test:
        assert num_not_run == 0
    else:
        assert 0 < num_not_run < len(suite)


def test_result_not_run_zero_when_all_success(suite):
    summary = suite.run()
    assert summary.session.results.get_num_not_run() == 0



def test_has_errors_or_failures(suite):
    suite[2].when_run.fail()
    suite[3].when_run.raise_exception()
    results = suite.run().session.results
    assert not results[0].has_errors_or_failures()
    assert results[2].has_errors_or_failures()
    assert results[3].has_errors_or_failures()


def test_has_skips(suite):
    suite[1].when_run.skip()
    results = suite.run().session.results
    assert not results[0].has_skips()
    assert results[1].has_skips()


def test_result_data_is_unique():

    class SampleTest(slash.Test):

        def test_1(self):
            pass

        def test_2(self):
            pass

    session = run_tests_assert_success(SampleTest)
    [result1, result2] = session.results
    assert result1.data is not result2.data


def test_result_test_garbage_collected(gc_marker):

    class SomeTest(slash.Test):
        def test_something(self):
            pass

    # we have to run another test at the end to make sure Slash's internal _last_test
    # doesn't refer to our test
    class OtherTest(slash.Test):
        def test_something(self):
            pass

    with slash.Session() as s:
        loader = slash.loader.Loader()
        tests = loader.get_runnables(SomeTest)
        # we use list(genexp) to prevent 't' from leaking
        marks = list(gc_marker.mark(t) for t in tests)
        session = run_tests_assert_success(tests + loader.get_runnables(OtherTest))
        del tests
    gc.collect()
    for mark in marks:
        assert mark.destroyed


def test_add_error_traceback_for_manually_added_errors(suite, suite_test):
    suite_test.append_line('slash.add_error("msg")')
    suite_test.expect_error()

    [result] = suite.run().get_all_results_for_test(suite_test)
    [err] = result.get_errors()
    assert err.traceback


class SessionResultTest(TestCase):

    def setUp(self):
        super(SessionResultTest, self).setUp()
        self.results = [
            Result() for _ in range(10)
        ]
        for r in self.results:
            r.mark_started()
        # one result with both errors and failures
        try:
            1 / 0
        except:
            self.results[1].add_error()
            self.results[1].add_failure()
            # one result with failure
            self.results[2].add_failure()
            # one result with error
            self.results[3].add_error()
            self.results[5].add_error()

        # one result will skip
        self.results[4].add_skip("Reason")

        # and one result will skip with error
        self.results[5].add_skip("Reason")
        num_finished = 7

        for result in self.results[:num_finished]:
            result.mark_finished()
        self.result = SessionResults(Session())
        for index, r in enumerate(self.results):
            self.result._results_dict[index] = r

    def test_counts(self):
        self.assertEquals(self.result.get_num_results(), len(self.results))
        self.assertEquals(self.result.get_num_successful(), 2)
        # errors take precedence over failures
        self.assertEquals(self.result.get_num_errors(), 3)
        self.assertEquals(self.result.get_num_skipped(), 2)
        self.assertEquals(self.result.get_num_failures(), 1)


@pytest.fixture(params=[GlobalResult, Result])
def result(request):
    return request.param()



import pytest

from .utils.suite_writer import Suite
from .utils.suite_writer.slash_run_result import Events


def test_run_suite_normal(suite):
    result = suite.run()
    assert result.session.results.get_num_successful() == len(suite)


def test_suite_events(suite):
    test1, test2 = suite[2], suite[3]
    result = suite.run()
    assert result.events['test_start', test1.id].is_before(result.events['test_start', test2.id])


def test_events_non_tuple():
    events = Events()
    events.add('a')
    assert 'a' in events
    assert 'b' not in events
    assert events.has_event('a')
    assert not events.has_event('b')


def test_events_tuple():
    events = Events()
    events.add('a', 'b', 'c')
    assert 'a' not in events
    assert 'b' not in events
    assert ('a', 'b', 'c') in events
    assert ('a',) not in events
    assert events.has_event('a', 'b', 'c')
    assert not events.has_event('a')





@pytest.fixture
def suite():
    s = Suite()
    for i in range(10):
        s.add_test()
    return s



import collections
import functools
import itertools

import pytest
import slash
from slash._compat import iteritems
from slash.core.scope_manager import ScopeManager
from slash.loader import Loader
from .utils import make_runnable_tests
from .utils.suite_writer import Suite


def test_requirement_mismatch_end_of_module():
    """Test that unmet requirements at end of file(module) still enable scope manager to detect the end and properly pop contextx"""

    suite = Suite()

    num_files = 3
    num_tests_per_file = 5

    for i in range(num_files):
        file1 = suite.add_file()

        for i in range(num_tests_per_file):
            file1.add_function_test()

        t = file1.add_function_test()
        t.add_decorator('slash.requires(lambda: False)')
        t.expect_skip()

    suite.run()


def test_scope_manager(dummy_fixture_store, scope_manager, tests_by_module):
    last_scopes = None
    for module_index, tests in enumerate(tests_by_module):
        for test_index, test in enumerate(tests):
            scope_manager.begin_test(test)
            assert dummy_fixture_store._scopes == ['session', 'module', 'test']
            expected = _increment_scope(
                last_scopes,
                test=1,
                module=1 if test_index == 0 else 0,
                session=1 if test_index == 0 and module_index == 0 else 0)
            assert dummy_fixture_store._scope_ids == expected
            # make sure the dict is copied
            assert expected is not dummy_fixture_store._scope_ids
            last_scopes = expected
            if test_index != len(tests) - 1:
                next_test = tests[test_index + 1]
                end_of_module = False
            elif module_index != len(tests_by_module) - 1:
                next_test = tests_by_module[module_index + 1][1]
                end_of_module = True
            else:
                next_test = None
            scope_manager.end_test(test, next_test=next_test, exc_info=(None, None, None))
            if next_test is None:
                assert dummy_fixture_store._scopes == []
            else:
                assert dummy_fixture_store._scopes == ['session', 'module']
            assert dummy_fixture_store._scope_ids == last_scopes

    scope_manager.flush_remaining_scopes()
    assert not dummy_fixture_store._scopes


@pytest.fixture
def scope_manager(dummy_fixture_store, forge):
    session = slash.Session()
    forge.replace_with(session, 'fixture_store', dummy_fixture_store)
    return ScopeManager(session)


@pytest.fixture
def dummy_fixture_store():
    return DummyFixtureStore()


@pytest.fixture
def tests_by_module():

    def test_func():
        pass

    num_modules = 5
    num_tests_per_module = 3
    returned = []

    with slash.Session():
        for module_index in range(num_modules):
            module_name = '__module_{0}'.format(module_index)
            returned.append([])
            for test_index in range(num_tests_per_module):
                [test] = make_runnable_tests(test_func)
                assert test.__slash__.module_name
                test.__slash__.module_name = module_name
                returned[-1].append(test)
    return returned


def _increment_scope(prev_scopes, **increments):
    if not prev_scopes:
        returned = {}
    else:
        returned = prev_scopes.copy()
    for key, value in iteritems(increments):
        if value == 0:
            continue
        if key not in returned:
            returned[key] = 0
        returned[key] += value
    return returned


class DummyFixtureStore(object):

    def __init__(self):
        super(DummyFixtureStore, self).__init__()
        self._scopes = []
        self._counters = collections.defaultdict(
            functools.partial(itertools.count, 1))
        self._scope_ids = {}

    def push_scope(self, scope):
        self._scopes.append(scope)
        self._scope_ids[scope] = next(self._counters[scope])

    def pop_scope(self, scope, in_failure, in_interruption):
        latest_scope = self._scopes.pop()
        assert latest_scope == scope



from __future__ import print_function

import sys
import traceback

from ..conf import config
from ..exceptions import SkipTest, INTERRUPTION_EXCEPTIONS
from ..ctx import context


def _debugger(debug_function_str, exc_info_transform=None):  # pragma: no cover
    module_name, function_name = debug_function_str.rsplit(".", 1)
    def debugger(exc_info):
        try:
            module = __import__(module_name, fromlist=[''])
        except ImportError:
            raise NotImplementedError() # pragma: no cover
        func = getattr(module, function_name)
        orig_exc_info = exc_info
        if exc_info_transform is not None:
            exc_info = exc_info_transform(exc_info)
        _notify_going_into_debugger(orig_exc_info)
        func(*exc_info)
    debugger.__name__ = debug_function_str
    return debugger

def _notify_going_into_debugger(exc_info):
    if context.session is not None:
        context.session.reporter.report_before_debugger(exc_info)
    else:
        print('\nException caught in debugger: {0}'.format(traceback.format_exception_only(exc_info[0], exc_info[1])[0].strip()))

def _only_tb(exc_info):  # pragma: no cover
    return (exc_info[2],)

def _tb_type_value(exc_info):  # pragma: no cover
    return (exc_info[2], exc_info[0], exc_info[1])

_KNOWN_DEBUGGERS = [
    # order is important here!
    _debugger("pudb.post_mortem", _tb_type_value),
    _debugger("ipdb.post_mortem", _only_tb),
    _debugger("pdb.post_mortem", _only_tb),
    ]


def debug_if_needed(exc_info=None):

    if not config.root.debug.enabled:
        return
    if exc_info is None:
        exc_info = sys.exc_info()
    if exc_info[0] is SkipTest and not config.root.debug.debug_skips:
        return
    if isinstance(exc_info[1], (SystemExit,) + INTERRUPTION_EXCEPTIONS):
        return

    launch_debugger(exc_info)

def launch_debugger(exc_info):
    for debug_func in _KNOWN_DEBUGGERS:
        try:
            debug_func(exc_info)
        except NotImplementedError:   # pragma: no cover
            continue
        else:
            break
    else:
        raise NotImplementedError("No debug function available")  # pragma: no cover



from .tagging import NO_TAGS

class RunnableTest(object):

    """
    This class is meant to serve as a base class to any test that can
    actually be executed by the Slash runner.
    """
    __slash__ = None

    def run(self):
        """
        This method is meant to be overriden by derived classes to actually
        perform the test logic
        """
        raise NotImplementedError()  # pragma: no cover

    def get_tags(self):
        return NO_TAGS

    def get_unmet_requirements(self):
        return [r for r in self.get_requirements() if not r.is_met()]

    def get_requirements(self):
        raise NotImplementedError() # pragma: no cover

    def get_required_fixture_objects(self):
        raise NotImplementedError() # pragma: no cover

    def __repr__(self):
        return '<Runnable test {0!r}>'.format(self.__slash__)



from .utils.suite_writer import Suite



def test_nofixtures(test_type):
    suite = Suite()
    test = suite.add_test(type=test_type)
    test.add_parameter_string('a=2, b=3')
    test.append_line('assert a == 2')
    test.append_line('assert b == 3')
    test.add_decorator('slash.nofixtures')
    assert suite.run().ok()



import arrow
import sys
import traceback

from .._compat import string_types, StringIO, iteritems
from ..exception_handling import is_exception_fatal
from ..exceptions import FAILURE_EXCEPTION_TYPES
from ..utils.traceback_utils import distill_traceback, distill_call_stack
from ..utils.formatter import Formatter


class Error(object):

    traceback = exception_type = exception = arg = _cached_detailed_traceback_str = None

    def __init__(self, msg=None, exc_info=None, frame_correction=0):
        super(Error, self).__init__()
        self.time = arrow.utcnow()
        self._fatal = False
        if msg is None and exc_info is not None:
            msg = traceback.format_exception_only(exc_info[0], exc_info[1])[0].strip()
        if not isinstance(msg, string_types):
            self.arg = msg
            msg = repr(msg)
        self.message = msg
        if exc_info is not None:
            self.exception_type, self.exception, tb = exc_info  # pylint: disable=unpacking-non-sequence
            self.traceback = distill_traceback(tb, frame_correction=frame_correction)
        else:
            self.traceback = distill_call_stack(frame_correction=frame_correction+4)

    def is_fatal(self):
        if self._fatal:
            return True
        return self.exception is not None and is_exception_fatal(self.exception)

    def mark_fatal(self):
        self._fatal = True
        return self

    def is_failure(self):
        return isinstance(self.exception, FAILURE_EXCEPTION_TYPES)

    @classmethod
    def capture_exception(cls):
        _, exc_value, _ = exc_info = sys.exc_info()
        if exc_value is None:
            return None
        cached = getattr(exc_value, "__slash_captured_error__", None)
        if cached is not None:
            return cached
        returned = exc_value.__slash_captured_error__ = cls(exc_info=exc_info)
        return returned

    @property
    def cause(self):
        if self.traceback is not None:
            return self.traceback.cause

    @property
    def filename(self):
        if self.traceback is not None:
            return self.traceback.cause.filename

    @property
    def lineno(self):
        if self.traceback is not None:
            return self.traceback.cause.lineno

    @property
    def func_name(self):
        if self.traceback is not None:
            return self.traceback.cause.func_name

    def __repr__(self):
        return self.message

    def get_detailed_traceback_str(self):
        if self._cached_detailed_traceback_str is None:
            stream = StringIO()
            f = Formatter(stream, indentation_string='  ')
            f.writeln("Traceback (most recent call last):")
            with f.indented():
                for frame in self.traceback.frames:
                    f.writeln('File "{f.filename}", line {f.lineno}, in {f.func_name}:'.format(f=frame))
                    with f.indented():
                        f.writeln('>', frame.code_line.strip() or '?')
                        with f.indented():
                            for title, vars in [('globals', frame.globals), ('locals', frame.locals)]:
                                for index, (var_name, var_repr) in enumerate(iteritems(vars)):
                                    if index == 0:
                                        f.writeln(title)
                                        f.indent()
                                    f.writeln(' - {0}: {1}'.format(var_name, var_repr['value']))
                            f.dedent()
            self._cached_detailed_traceback_str = stream.getvalue()

        return self._cached_detailed_traceback_str

    def get_detailed_str(self):
        return '{0}*** {1}'.format(
            self.get_detailed_traceback_str(), self)




import collections
import itertools
from numbers import Number

from ..._compat import iteritems, OrderedDict, string_types, imap, izip, reduce, xrange
from ...exceptions import FixtureException
from ...utils.python import getargspec
from .parameters import Parametrization, get_parametrizations
from .utils import nofixtures

_PRINTABLE_TYPES = (Number,) + string_types


class VariationFactory(object):

    """Helper class to produce variations, while properly naming the needed fixtures to help identifying tests
    """

    def __init__(self, fixture_store):
        super(VariationFactory, self).__init__()
        self._store = fixture_store
        self._needed_fixtures = list(fixture_store.iter_autouse_fixtures_in_namespace())
        self._named_fixture_ids = OrderedDict()
        self._known_value_strings = collections.defaultdict(dict)

    def add_needed_fixture_id(self, fixture_id):
        self._needed_fixtures.append(self._store.get_fixture_by_id(fixture_id))

    def add_needed_fixtures_from_method(self, method):
        self._add_needed_fixtures_from_function(method, is_method=True)

    def add_needed_fixtures_from_function(self, func):
        self._add_needed_fixtures_from_function(func, is_method=False)

    def _add_needed_fixtures_from_function(self, func, is_method):

        if isinstance(func, tuple):
            namespace, func = func
        else:
            namespace = None

        if nofixtures.is_marked(func):
            return

        arg_names = getargspec(func).args[1 if is_method else 0:]

        parametrizations = {}
        for param in get_parametrizations(func):
            # make sure the parametrization is in the store
            self._store.ensure_known_parametrization(param)
            for name in param.names:
                parametrizations[name] = param

        for arg_name in arg_names:
            fixture = parametrizations.get(arg_name, None)
            if fixture is None:
                try:
                    fixture = self._store.get_fixture_by_name(arg_name)
                except FixtureException as e:
                    raise type(e)('Loading {0.__code__.co_filename}:{0.__name__}: {1}'.format(func, e))


            self._needed_fixtures.append(fixture)
            if namespace is not None:
                arg_name = '{0}:{1}'.format(namespace, arg_name)
            self._named_fixture_ids[fixture.info.id] = arg_name

    def iter_variations(self):
        param_ids = list(reduce(set.union, imap(self._store.get_all_needed_parametrization_ids, self._needed_fixtures), set()))
        parametrizations = [self._store.get_fixture_by_id(param_id) for param_id in param_ids]
        if not param_ids:
            yield Variation(self._store)
            return
        for value_indices in itertools.product(*(xrange(len(p.values)) for p in parametrizations)):
            yield self._build_variation(parametrizations, value_indices)

    def _build_variation(self, parametrizations, value_indices):
        param_value_indices = dict((p.info.id, param_index)
                              for p, param_index in izip(parametrizations, value_indices))
        return Variation(self._store, param_value_indices, representation=self._build_representation(param_value_indices))

    def _build_representation(self, param_value_indices):

        if not self._named_fixture_ids:
            return None

        parts = []
        for fixture_id, fixture_name in iteritems(self._named_fixture_ids):
            value_str = self._get_fixture_value_string(fixture_id, fixture_name, param_value_indices)
            parts.append('{0}={1}'.format(fixture_name, value_str))
        return ', '.join(parts)

    def _get_fixture_value_string(self, fixture_id, fixture_name, param_value_indices):
        fixture = self._store.get_fixture_by_id(fixture_id)
        needed_ids = self._store.get_all_needed_parametrization_ids(fixture)
        combination = frozenset((param_id, param_value_indices[param_id]) for param_id in needed_ids)
        value_str = self._known_value_strings[fixture_id].get(combination)
        if value_str is None:
            if isinstance(fixture, Parametrization) and self._is_printable(fixture.values[param_value_indices[fixture_id]]):
                value_str = str(fixture.values[param_value_indices[fixture_id]])
            else:
                value_str = '{0}{1}'.format(fixture_name, len(self._known_value_strings[fixture_id]))
                self._known_value_strings[fixture_id][combination] = value_str
        return value_str

    def _is_printable(self, value):
        return isinstance(value, _PRINTABLE_TYPES) and not (isinstance(value, string_types) and '/' in value)


class Variation(object):

    """Represents a single variation of parameter points. A variation is merely a mapping of fixture ids to their values.
    This mostly applies for parametrization fixtures. The other fixtures follow since they are either constant
    or indirectly depend on parametrization"""

    def __init__(self, store, param_value_indices=None, representation=None):
        super(Variation, self).__init__()
        self._store = store
        if param_value_indices is None:
            param_value_indices = {}
        self.param_value_indices = param_value_indices
        self.representation = representation

    def has_value_for_parameter(self, param_id):
        return param_id in self.param_value_indices

    def get_param_value(self, param_id):
        return self._store.get_fixture_by_id(param_id).values[self.param_value_indices[param_id]]

    def __eq__(self, other):
        if isinstance(other, Variation):
            other = other.param_value_indices
        if not isinstance(other, dict):
            return NotImplemented
        return self.param_value_indices == other

    def __ne__(self, other):
        return not (self == other)  # pylint: disable=superfluous-parens

    def __repr__(self):
        return 'Variation({0})'.format(', '.join('{0}={1}'.format(key, value) for key, value in iteritems(self.param_value_indices)))

    def __nonzero__(self):
        return bool(self.param_value_indices)

    __bool__ = __nonzero__



import itertools

import gossip
import pytest
import slash

from .utils import CustomException, run_tests_in_session


def test_scope_management_with_hook_error_test_start(suite, suite_test):
    assert len(suite.files) > 1
    assert len(suite) > 1

    @gossip.register('slash.test_start')
    def hook():
        raise CustomException()

    for test in suite:
        test.expect_error()
    summary = suite.run()
    for res in summary.session.results.iter_test_results():
        [err] = res.get_errors()
        assert err.exception_type is CustomException

def test_scope_management_with_hook_error_test_end():
    """test_end errors are fatal, so the session abruptly stops. We just make sure we get the exception and that at least one test runs"""
    events = []

    gossip.register('slash.test_end')(CustomException.do_raise)

    @slash.parametrize('param', range(10))
    def test_something(param):
        events.append('test is running!')

    with slash.Session() as session:
        tests = slash.loader.Loader().get_runnables(test_something)
        assert tests
        with pytest.raises(CustomException) as caught:
            run_tests_in_session(tests, session=session)

    assert len(events) == 1
    assert len(tests) > 1



import copy

import pytest
import slash

from .utils import run_tests_assert_success
from .utils.suite_writer import Suite


def test_test_parametrization(test_type):
    suite = Suite()
    num_values1 = 3
    num_values2 = 5
    test = suite.add_test(type=test_type)
    test.add_parameter(num_values=num_values1)
    test.add_parameter(num_values=num_values2)
    summary = suite.run()
    assert len(summary.get_all_results_for_test(test)) == num_values1 * num_values2


def test_parameters_toggle():

    @slash.parameters.toggle('param')
    def test_example(param):
        _set('param', param)

    session = run_tests_assert_success(test_example)

    assert [False, True] == sorted(result.data['param'] for result in session.results)


def test_dict_parameter_values():

    values = [{'value': 1}, {'value': 2}]

    @slash.parametrize('param', copy.deepcopy(values))
    def test_example(param):
        _set('param', param)

    session = run_tests_assert_success(test_example)
    assert values == sorted((result.data['param'] for result in session.results), key=values.index)


def test_duplicate_parameters():

    values = ["a", "b", "a"]

    @slash.parametrize('param', copy.deepcopy(values))
    def test_example(param):
        _set('param', param)

    session = run_tests_assert_success(test_example)
    assert sorted(values) == sorted(result.data['param'] for result in session.results)



def test_before_after_parameters(cartesian):

    class Parameterized(slash.Test):

        @slash.parameters.iterate(a=cartesian.before_a.make_set())
        def before(self, a):
            _set("before_a", a)

        @slash.parameters.iterate(b=cartesian.b.make_set(), c=cartesian.c.make_set())
        def test(self, b, c):
            _set("b", b)
            _set("c", c)

        @slash.parameters.iterate(d=cartesian.after_d.make_set())
        def after(self, d):
            _set("after_d", d)

    session = run_tests_assert_success(Parameterized)
    assert len(session.results) == len(cartesian)
    cartesian.check(result.data for result in session.results)


@pytest.mark.parametrize('with_override', [True, False])
def test_before_parameters_inheritence(cartesian, with_override):

    class BaseTest(slash.Test):

        @slash.parameters.iterate(a=cartesian.before_1_a.make_set())
        def before(self, a):
            _set("before_1_a", a)

    class DerivedTest(BaseTest):

        @slash.parameters.iterate(a=cartesian.before_2_a.make_set(), b=cartesian.before_2_b.make_set())
        def before(self, a, b):
            if with_override:
                super(DerivedTest, self).before(a=a)
            else:
                super(DerivedTest, self).before()
            _set("before_2_a", a)
            _set("before_2_b", b)

        def test(self):
            pass

    session = run_tests_assert_success(DerivedTest)
    assert len(session.results) == len(cartesian)
    if with_override:
        cartesian.assign_all(
            source_name='before_2_a', target_name='before_1_a')
    cartesian.check(result.data for result in session.results)


def test_parametrization_tuples():

    @slash.parametrize(('a', 'b'), [(1, 2), (11, 22)])
    @slash.parametrize('c', [3, 33])
    def test_something(a, b, c):
        _set("params", (a, b, c))

    session = run_tests_assert_success(test_something)
    results = [result.data['params'] for result in session.results.iter_test_results()]
    expected = set([
        (1, 2, 3), (1, 2, 33), (11, 22, 3), (11, 22, 33)
        ])
    assert len(expected) == len(results)
    assert expected == set(results)

def test_parametrization_tuples_invalid_length():

    with pytest.raises(RuntimeError) as caught:
        @slash.parametrize(('a', 'b'), [(1, 2), (1,), (11, 22)])
        def test_something(a, b, c):
            pass
    assert 'Invalid parametrization value' in str(caught.value)
    assert 'invalid length' in str(caught.value)


def test_parametrization_tuples_invalid_type():

    with pytest.raises(RuntimeError) as caught:
        @slash.parametrize(('a', 'b'), [(1, 2), 200, (11, 22)])
        def test_something(a, b, c):
            pass
    assert 'Invalid parametrization value' in str(caught.value)
    assert 'expected sequence' in str(caught.value)



def _set(param, value):
    data = slash.session.results.current.data
    assert param not in data
    data[param] = value



from sentinels import NOTHING
from numbers import Number

from ..._compat import itervalues
from ...exceptions import UnknownFixtures


class Namespace(object):

    def __init__(self, store, parent=None):
        super(Namespace, self).__init__()
        self._store = store
        self._parent = parent
        self._fixture_names = {}

    def iter_fixtures(self):
        while self is not None:
            for fixture_id in itervalues(self._fixture_names):
                yield self._store.get_fixture_by_id(fixture_id)
            self = self._parent

    def __repr__(self):
        return 'Fixtures: {0}'.format(', '.join(self._iter_fixture_names()) or '**None**')

    def _iter_fixture_names(self):
        while self is not None:
            for k in self._fixture_names:
                yield k
            self = self._parent

    def get_fixture_by_name(self, name):
        while self is not None:
            fixture_id = self._fixture_names.get(name, NOTHING)
            if fixture_id is NOTHING:
                self = self._parent
                continue
            return self._store.get_fixture_by_id(fixture_id)
        raise UnknownFixtures('Fixture {0!r} not found!'.format(name))

    def add_name(self, name, fixture_id):
        assert isinstance(fixture_id, Number)
        assert name not in self._fixture_names or self._fixture_names[
            name] == fixture_id
        self._fixture_names[name] = fixture_id



import sys

import emport

from slash.core.error import Error


def test_traceback_line_numbers(tmpdir):
    filename = tmpdir.join('filename.py')

    with filename.open('w') as f:
        f.write('''from contextlib import contextmanager
def f():
    with context():
        a = 1
        b = 2
        g()
        c = 3
        d = 4

def g():
    1/0

@contextmanager
def context():
    yield
''')

    mod = emport.import_file(str(filename))
    try:
        mod.f()
    except ZeroDivisionError:
        err = Error(exc_info=sys.exc_info())
    else:
        assert False, 'did not fail'

    assert err.traceback.frames[-2].lineno == 6



import abc

class PluginInterface(object):
    """
    This class represents the base interface needed from plugin classes.
    """
    __metaclass__ = abc.ABCMeta

    def activate(self):
        """
        Called when the plugin is activated
        """
        pass

    def get_config(self):
        """
        Optional: should return a dictionary or a confetti object which will be placed under
        ``slash.config.plugin_config.<plugin_name>``
        """
        pass

    def deactivate(self):
        """
        Called when the plugin is deactivated

        .. note:: this method might not be called in practice, since it is not guaranteed that plugins are always
          deactivated upon process termination. The intention here is to make plugins friendlier to cases
          in which multiple sessions get established one after another, each with a different set of plugins.
        """
        pass

    def configure_argument_parser(self, parser):
        """
        Gives a chance to the plugin to add options received from command-line
        """
        pass

    def configure_from_parsed_args(self, args):
        """
        Called after successful parsing of command-line arguments
        """
        pass

    def get_description(self):
        """
        Retrieves a quick description for this plugin, mostly used in command-line help or online documentation.
        It is not mandatory to override this method.
        """
        return None

    def get_name(self):
        """
        Returns the name of the plugin class. This name is used to register, disable and address
        the plugin during runtime.

        Note that the command-line switches (``--with-...``) are derived from this name.

        Any implemented plugin must override this method.
        """
        raise NotImplementedError() # pragma: no cover
